/*-------------------------------------------------------------------------*/
/* Metodos de Pesquisa e ordenacao                                         */
/* 2001 - Joao Luiz Marciano                                               */
/*-------------------------------------------------------------------------*/

#include <time.h>
#include <math.h>
#include <conio.h>
#include <stdio.h>
#include <stdlib.h>

#define TRUE      1
#define FALSE     0
#define N     50000

int a[N];                 // vetor de numeros para ordenacao

/*--------------------------------------------------------------------------*/
/* Pesquisa sequencial                                                      */
/*--------------------------------------------------------------------------*/

int PesquisaSequencial(int Li,int Ls,int Chave) {

        int i;

        for (i=Li;i<Ls;i++)
                if (a[i] == Chave)
                        return i;
        return -1;
}

/*--------------------------------------------------------------------------*/
/* Pesquisa binaria                                                         */
/*--------------------------------------------------------------------------*/

int PesquisaBinaria(int Li,int Ls,int Chave) {
        int j,m;
        div_t s;

        for (;;) {

                s = div(Li + Ls,2);
                m = s.quot;

                if (Chave < a[m])
                        Ls = m - 1;
                else if (Chave > a[m])
                        Li = m + 1;
                     else return m;

                if (Li > Ls)
                        return -1;
        }
}

/*--------------------------------------------------------------------------*/
/* Ordenacao por selecao                                                    */
/*--------------------------------------------------------------------------*/

void Selecao(void) {
        int i,j,min,t;

        for (i=0;i<N;i++) {

                min=i;
                for (j=i+1;j<N;j++)
                        if (a[j] < a[min]) min=j;

                t = a[min];
                a[min] = a[i];
                a[i] = t;
        }
}

/*--------------------------------------------------------------------------*/
/* Ordenacao por insercao                                                   */
/*--------------------------------------------------------------------------*/

void Insercao(void) {
        int i,j,v;

        for (i=1;i<N;i++) {

                v=a[i];
                j=i;
                while ((j > 0) && (a[j-1] > v)) {

                        a[j]=a[j-1]; // estas duas instrucoes podem ser
                        j--;         // substituidas por a[j]=a[--j];
                }
                a[j]=v;
        }
}

/*--------------------------------------------------------------------------*/
/* Bubblesort                                                               */
/*--------------------------------------------------------------------------*/

void Bolha(void) {
        int i,j,t;

        for (i=N-1;i>=0;i--)
                for (j=2;j<=i;j++)
                        if (a[j-1] > a[j]) {

                                t=a[j-1];
                                a[j-1]=a[j];
                                a[j]=t;
                        }
}

/*--------------------------------------------------------------------------*/
/* Shellsort                                                                */
/*--------------------------------------------------------------------------*/

void Shell(void) {
        int i,j,h,v;
        div_t s;

        /* determinacao do valor de h */

        h=1;

        do {
                h = 3 * h + 1;
        } while (h <= N);

        do {

                s = div(h,3);
                h = s.quot;

                for (i=h+1;i<N;i++) {

                        v=a[i];
                        j=i;

                        while (a[j-h] > v) {

                                a[j] = a[j-h];
                                j-=h;

                                if (j<=h) break;

                        }

                        a[j] = v;
                }
        } while (h!=1);
}

/*--------------------------------------------------------------------------*/
/* Quicksort                                                                */
/*--------------------------------------------------------------------------*/

void Quick(int e,int d) {

        int i,j,v,t;

        if (d > e) {

                v=a[d];
                i=e-1;
                j=d;

                do {

                        do {
                                i++;
                        } while (a[i] < v);

                        do {
                                j--;
                        } while (a[j] > v);

                        t=a[i];
                        a[i]=a[j];
                        a[j]=t;

                } while (j > i);

                a[j]=a[i];
                a[i]=a[d];
                a[d]=t;

                Quick(e,i-1);
                Quick(i+1,d);
        }
}


/*--------------------------------------------------------------------------*/
/* Impressao                                                                */
/*--------------------------------------------------------------------------*/

void Imprime(void) {

        int i;

        for (i=0;i<N;i++)
                printf("%d\n",a[i]);
}

/*--------------------------------------------------------------------------*/
/* Funcao principal                                                         */
/*--------------------------------------------------------------------------*/

void main(void) {

        int i,j;
        char opcao[10];
        time_t s,t;    // estas estruturas cronometram o tempo de execucao das ordenacoes

        clrscr();

        // inicializa gerador de numeros aleatorios

        randomize();

        // ordenacao

        do {
                printf("Selecione o metodo de ORDENACAO a ser usado.\n");
                printf("Serao gerados N=50000 numeros inteiros para ordenar.\n");
                printf("Ao termino da ordenacao, sera mostrado o tempo gasto para faze-la.\n");
                printf("1=Selecao/2=Insercao/3=Bolha/4=Shellsort/5=Quicksort: (<enter>=fim das ordenacoes)");
                gets(opcao);

                if (opcao[0] != '\0') {
                        j = atoi(opcao);

                        for (i=0;i<N;i++)
                                a[i] = random(N);

                        switch(j) {
                                case 1: s = time(NULL);
                                        Selecao();
                                        t = time(NULL);
                                        printf("Selecao gastou %ld segundos\n",t-s);
                                        break;

                                case 2: s = time(NULL);
                                        Insercao();
                                        t = time(NULL);
                                        printf("Insercao gastou %ld segundos\n",t-s);
                                        break;

                                case 3: s = time(NULL);
                                        Bolha();
                                        t = time(NULL);
                                        printf("Bolha gastou %ld segundos\n",t-s);
                                        break;

                                case 4: s = time(NULL);
                                        Shell();
                                        t = time(NULL);
                                        printf("Shell gastou %ld segundos\n",t-s);
                                        break;

                                case 5: s = time(NULL);
                                        Quick(0,N-1);
                                        t = time(NULL);
                                        printf("Quick gastou %ld segundos\n",t-s);
                                        break;

                        }
                }

        }
        while (opcao[0] != '\0');
}

