/*----------------------------------------------------------------------------*/
/* Programa para a insercao, pesquisa e remocao de nos em arvores balanceadas */
/*                                                                            */
/* Adaptado de Wirth, Algoritmos e estruturas de dados                        */
/*----------------------------------------------------------------------------*/

#include <conio.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define TRUE  1
#define FALSE 0

typedef struct No* noptr;

struct No {
        int valor;
        int contador;
        int bal;
        noptr esquerda;
        noptr direita;
};

/*----------------------------------------------------------------------------*/
/* Funcao de impressao da arvore                                              */
/*----------------------------------------------------------------------------*/

void ImprimeArvore (noptr t, int h) {
        int i;

        if (t) {

                ImprimeArvore(t->esquerda,h+1);
                for (i=0;i<h;i++) printf(" ");
                printf("%d com %d ocorrencia(s)\n",t->valor,t->contador);
                ImprimeArvore(t->direita,h+1);
        }
}

/*----------------------------------------------------------------------------*/
/* Funcao de montagem da arvore                                               */
/*----------------------------------------------------------------------------*/

void Pesquisa (int x,noptr* p, int *h) {
noptr p1,p2;

        if (*p == NULL) {

                *h = TRUE;
                *p = (noptr) malloc(sizeof(struct No));
                (*p)->valor    = x;
                (*p)->contador = 1;
                (*p)->esquerda = NULL;
                (*p)->direita  = NULL;
                (*p)->bal = 0;
        }

        else if (x < (*p)->valor) {
                Pesquisa(x,&((*p)->esquerda),h);
                if (*h) {                             /* o ramo esq cresceu */
                        switch ((*p)->bal) {
                                case 1: (*p)->bal = 0;
                                        *h = FALSE;
                                        break;
                                case 0: (*p)->bal = -1;
                                        break;
                                case -1:p1 = (*p)->esquerda;  /* rebalanceamento */
                                        if (p1->bal == -1) {  /* rotacao simples esq */
                                                (*p)->esquerda = p1->direita;
                                                p1->direita = *p;
                                                (*p)->bal = 0;
                                                *p = p1;
                                        }
                                        else {  /* rotacao dupla esq-dir */
                                                p2 = p1->direita;
                                                p1->direita = p2->esquerda;
                                                p2->esquerda = p1;
                                                (*p)->esquerda = p2->direita;
                                                p2->direita = *p;
                                                if (p2->bal == -1)
                                                        (*p)->bal = 1;
                                                else
                                                        (*p)->bal = 0;
                                                if (p2->bal == 1)
                                                        p1->bal = -1;
                                                else
                                                        p1->bal = 0;
                                                *p = p2;
                                        }
                                        (*p)->bal = 0;
                                        *h = FALSE;
                                        break;
                        }

                }
        }
        else if (x > (*p)->valor) {
                Pesquisa(x,&((*p)->direita),h);
                if (*h) {                             /* o ramo dir cresceu */
                        switch((*p)->bal) {
                                case -1:(*p)->bal = 0;
                                        *h = FALSE;
                                        break;
                                case 0: (*p)->bal = 1;
                                        break;
                                case 1: p1 = (*p)->direita; /* rebalanceamento */
                                        if (p1->bal == 1) { /* rotacao simples dir */
                                                (*p)->direita = p1->esquerda;
                                                p1->esquerda = *p;
                                                (*p)->bal = 0;
                                                *p = p1;
                                        }
                                        else { /* rotacao dupla dir-esq */
                                                p2 = p1->esquerda;
                                                p1->esquerda = p2->direita;
                                                p2->direita = p1;
                                                (*p)->direita = p2->esquerda;
                                                p2->esquerda = *p;
                                                if (p2->bal == 1)
                                                        (*p)->bal = -1;
                                                else
                                                        (*p)->bal = 0;
                                                if (p2->bal == -1)
                                                        p1->bal = 1;
                                                else
                                                        p1->bal = 0;
                                                *p = p2;
                                        }
                                        (*p)->bal = 0;
                                        *h = FALSE;
                                        break;
                        }
                }
        }
        else (*p)->contador++;
}

/*----------------------------------------------------------------------------*/
/* Funcao de balanceamento a esquerda                                         */
/*----------------------------------------------------------------------------*/

void balanceE(noptr* p,int *h){
        noptr p1,p2;
        int b1,b2;

        /* o ramo esquerdo diminuiu */

        switch ((*p)->bal) {
                case -1: (*p)->bal = 0;
                          break;
                case 0:  (*p)->bal = 1;
                          *h = FALSE;
                          break;
                case 1: /* rebalanceamento */
                          p1 = (*p)->direita;
                          b1 = p1->bal;
                          if (b1 >= 0) {        /* rotacao simples dir */
                                (*p)->direita = p1->esquerda;
                                p1->esquerda = *p;
                                if (b1 == 0)  {
                                        (*p)->bal = 1;
                                        p1->bal = -1;
                                        *h = FALSE;
                                }
                                else {
                                        (*p)->bal = 0;
                                        p1->bal = 0;
                                }
                                *p = p1;
                          }
                          else { /* rotacao dupla dir-esq */
                                p2 = p1->esquerda;
                                b2 = p2->bal;
                                p1->esquerda = p2->direita;
                                p2->direita = p1;
                                (*p)->direita = p2->esquerda;
                                p2->esquerda = *p;
                                if (b2 == 1)
                                        (*p)->bal = -1;
                                else
                                        (*p)->bal = 0;
                                if (b2 == -1)
                                        p1->bal = 1;
                                else
                                        p1->bal = 0;
                                *p = p2;
                                p2->bal = 0;
                          }
                          break;
        }
}

/*----------------------------------------------------------------------------*/
/* Funcao de balanceamento a direita                                          */
/*----------------------------------------------------------------------------*/

void balanceD(noptr *p,int *h) {
        noptr p1,p2;
        int b1,b2;

        /* o ramo dir diminuiu */

        switch ((*p)->bal) {
                case 1: (*p)->bal = 0;
                          break;
                case 0:  (*p)->bal = -1;
                          *h = FALSE;
                          break;
                case -1: /* rebalanceamento */
                          p1 = (*p)->esquerda;
                          b1 = p1->bal;
                          if (b1 <= 0) {        /* rotacao simples esq */
                                (*p)->esquerda = p1->direita;
                                p1->direita = *p;
                                if (b1 == 0)  {
                                        (*p)->bal = -1;
                                        p1->bal = 1;
                                        *h = FALSE;
                                }
                                else {
                                        (*p)->bal = 0;
                                        p1->bal = 0;
                                }
                                *p = p1;
                          }
                          else { /* rotacao dupla esq-dir */
                                p2 = p1->direita;
                                b2 = p2->bal;
                                p1->direita = p2->esquerda;
                                p2->esquerda = p1;
                                (*p)->esquerda = p2->direita;
                                p2->direita = *p;
                                if (b2 == -1)
                                        (*p)->bal = 1;
                                else
                                        (*p)->bal = 0;
                                if (b2 == 1)
                                        p1->bal = -1;
                                else
                                        p1->bal = 0;
                                *p = p2;
                                p2->bal = 0;
                          }
                          break;
        }
}


/*----------------------------------------------------------------------------*/
/* Funcao auxiliar de remocao                                                */
/*----------------------------------------------------------------------------*/

void del (noptr* q, noptr* r,int *h) {

        if ((*r)->direita != NULL) {
                del(q,&((*r)->direita),h);
                if (*h) balanceD(r,h);
        }
        else {
                (*q)->valor = (*r)->valor;
                (*q)->contador = (*r)->contador;
                *q = *r;
                *r = (*r)->esquerda;
                *h = TRUE;
        }
}

/*----------------------------------------------------------------------------*/
/* Funcao de remocao                                                          */
/*----------------------------------------------------------------------------*/

void Delete (int x, noptr* p, int *h) {

        noptr q;

        if (*p == NULL) {
                printf("O no %d nao se encontra na arvore\n",x);
        }        
        else if (x < (*p)->valor) {
                Delete(x,&((*p)->esquerda),h);
                if (*h) balanceE(p,h);
             }
             else if (x > (*p)->valor) {
                        Delete(x,&((*p)->direita),h);
                        if (*h) balanceD(p,h);
                  }
                  else { /* remover p */
                                q = *p;
                                if (q->direita == NULL) {
                                        *p = q->esquerda;
                                        *h = TRUE;
                                }
                                else if (q->esquerda == NULL) {
                                        *p = q->direita;
                                        *h = TRUE;
                                     }
                                     else {
                                        del(&q,&(q->esquerda),h);
                                        if (*h) balanceE(p,h);
                                     }
                  }
}

/*----------------------------------------------------------------------------*/
/*                                Funcao principal                            */
/*----------------------------------------------------------------------------*/

void main() {
        int i,N,valor,h;
        noptr Raiz;
        char str[10];

        clrscr();

        /* constroi a arvore */

        Raiz = NULL;
        printf("Indique o numero de nos a ler (<enter>=fim): ");

        gets(str);

        if (str[0] != '\0') {

                sscanf(str,"%d",&N);

                for (i=1;i<=N;i++){
                        printf("Forneca o no %d: ",i);
                        gets(str);
                        valor = atoi(str);
                        Pesquisa(valor,&Raiz,&h);
                }

                ImprimeArvore(Raiz,0);

                /* remove nos da arvore */

                do {
                        printf("Forneca o no a remover (<enter> para encerrar): ");
                        gets(str);

                        if (str[0] != '\0') {
                                valor = atoi(str);
                                Delete(valor,&Raiz,&h);
                                ImprimeArvore(Raiz,0);
                        }
                } while (str[0] != '\0');
        }
}

